@@ -50,16 +50,22 @@ child_specs() ->
     YawsLog = {yaws_log, {yaws_log, start_link, []},
                permanent, 5000, worker, [yaws_log]},
 
+    YawsTrace = {yaws_trace, {yaws_trace, start_link, []},
+                 permanent, 5000, worker, [yaws_trace]},
+
     YawsServArgs = [_Env = get_app_args()],
     YawsServ = {yaws_server, {yaws_server, start_link, YawsServArgs},
                 permanent, 120000, worker, [yaws_server]},
 
+    YawsDavLock = {yaws_davlock, {yaws_davlock, start_link, []},
+                 permanent, 5000, worker, [yaws_davlock]},
+
     %% and this guy will restart auxiliary procs that can fail
     Sup = {yaws_sup_restarts,
            {yaws_sup_restarts, start_link, []},
            transient, infinity, supervisor, [yaws_sup_restarts]},
 
-    [YawsLog, YawsServ, Sup].
+    [YawsLog, YawsTrace, YawsServ, YawsDavLock, Sup].
 
 %%----------------------------------------------------------------------
 %%----------------------------------------------------------------------
